from __future__ import annotations

import sys
import types
import warnings
from functools import partial
from typing import (
    Any,
    Callable,
    ClassVar,
    Coroutine,
    Generic,
    Iterable,
    List,
    MutableMapping,
    TypeVar,
)

from deprecated_params import deprecated_params
from frozenlist import FrozenList
from propcache import cached_property

if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec
else:
    from typing_extensions import Concatenate, ParamSpec

T = TypeVar("T")
P = ParamSpec("P")
OwnerT = TypeVar("OwnerT")

AsyncFunction = Callable[P, Coroutine[Any, Any, T]]


# EventWrapper is on par with aiosignal instead of being it's subclass
# This saves a couple of steps for many of our subclasses...


class EventWrapper(FrozenList[AsyncFunction[P, T]]):
    """A wrapper class for making a callback function that carries a few more methods than aiosignal has."""

    __slots__ = ("_owner",)

    def __init__(
        self,
        items: List[AsyncFunction[P, T]] | Iterable[AsyncFunction[P, T]] | None = None,
        /,
        owner=None,
    ):
        super().__init__(items)
        self._owner = owner

    def __call__(self, func: AsyncFunction[P, T]):
        """appends a callback function to the event, returns the same function for futher use elsewhere...
        this is equivilent to calling the `append()` method::

            from aiocallback import EventWrapper

            custom_event = EventWrapper()
            @custom_event
            async def on_event():
                ...

        """
        self.append(func)
        return func

    # Typehint our signal so that pyright can see
    # the arguments that need to be passed
    async def send(self, *args: P.args, **kw: P.kwargs) -> None:
        if not self.frozen:
            raise RuntimeError("Cannot send non-frozen events.")

        for receiver in self:
            await receiver(*args, **kw)  # type: ignore

    def __repr__(self):
        return "<{} owner={}, frozen={}, {!r}>".format(
            self.__class__.__name__, self._owner, self.frozen, list(self)
        )


class SelfEventWrapper(EventWrapper[P, T]):
    """A wrapper class for making an owner object sendable with all the events"""

    def __init__(self, items=None, /, owner=None):
        super().__init__(items, owner)

    async def send(self, *args: P.args, **kwargs: P.kwargs) -> None:
        return await super().send(self._owner, *args, **kwargs)  # type: ignore


@deprecated_params(
    ["abstract"],
    "abstract keyword is deprecated & it's functionality has been removed, removal of warning planned in 0.1.7",
)
class event(Generic[OwnerT, P, T]):
    """A Couroutine Based implementation of an asynchronous callback object. 
    This object is a replacement for aiosignal. with easier configuration options...
    
    abstract: `bool` inner function upon being called is considered abstract... \
        if `true` inner custom function will not be called with the `send()` method and it \
        will be considered as nothing but typehinting.
    """

    __slots__ = ("func", "name", "_wrapper")

    _wrapper: EventWrapper[P, T]

    def __init__(self, func: AsyncFunction[Concatenate[OwnerT, P], T], **kw) -> None:
        self.func = func
        self.name = func.__name__

        if hasattr(func, "__isabstractmethod__"):
            warnings.warn(
                "using an abc.abstractmethod wrapper with an event is discouraged",
                UserWarning,
                2,
            )

    # __doc__ couldn't be made into a slot
    # so we had to come up with an alternative method
    @cached_property
    def __doc__(self):
        return self.func.__doc__

    def __wrapper_init__(self, owner: OwnerT):
        """A Special dunder method for initalizing an Event Wrapper"""
        self._wrapper = EventWrapper(owner=owner)
        return self._wrapper

    # Turns the event into a descriptor variable
    # SEE: https://docs.python.org/3/howto/descriptor.html
    # To sumarize it gets called during `__new__` which means that the wrapper
    # will always attempt to be inbounds...
    def __set_name__(self, owner: OwnerT, name: str):
        self.__wrapper_init__(owner)
        self.name = name

    # inner _event_cache is removed because using slots on the descriptor is faster

    def __get__(self, inst: OwnerT, owner) -> EventWrapper[P, T]:
        # if for some reason we did not obtain this during __new__...
        if not hasattr(self, "_wrapper"):
            self.__wrapper_init__(owner)
        return self._wrapper

    __class_getitem__ = classmethod(types.GenericAlias)  # type:ignore


class subclassevent(event[OwnerT, P, T]):
    """Passes the context class to the member descriptor"""

    def __wrapper_init__(self, owner: OwnerT):
        self._wrapper = EventWrapper((partial(self.func, owner),), owner)
        return self._wrapper

    def __get__(self, inst: OwnerT, owner) -> EventWrapper[P, T]:
        # Incase the user's object does not have a base property to use...
        if not hasattr(self, "_wrapper") or self._wrapper._owner != inst:
            # Call the instance instead so that the instance is called with the event
            self.__wrapper_init__(inst)
        return self._wrapper


class contextevent(event[OwnerT, P, T]):
    """Sends the class holding the event through each of the callbacks made except for the wrapper itself."""

    _wrapper: SelfEventWrapper[P, T]

    def __wrapper_init__(self, owner: OwnerT):
        self._wrapper = SelfEventWrapper(owner=owner)
        return self._wrapper

    def __get__(self, inst: OwnerT, owner) -> SelfEventWrapper[P, T]:
        # Incase the user's object does not have a base property to use...
        if not hasattr(self, "_wrapper") or self._wrapper._owner != inst:
            self.__wrapper_init__(inst)
        return self._wrapper


class subcontextevent(contextevent):
    """sends the class holding the event as an instance through all the callbacks made including the inner wrapper
    being utilized."""

    def __wrapper_init__(self, owner):
        self._wrapper = SelfEventWrapper((self.func,), owner)
        return self._wrapper


class subclasscontextevent(subcontextevent):
    def __init__(self, func, **kw):
        warnings.warn(
            "subclasscontextevent event has been"
            " renamed to subcontextevent (to try and lessen confusion),"
            " subclasscontextevent will be removed in 0.1.7",
            DeprecationWarning,
            2,
        )
        super().__init__(func, **kw)


# Inspired by PEP 3115's example
class event_table(dict):
    __slots__ = ("events",)

    def __init__(self):
        self["_events"] = {}

    def __setitem__(self, key: str, value):
        # if the key is not already defined, add it to the
        # list of keys.
        if key not in self:
            # see if were either an event or context event.
            if isinstance(value, (event, contextevent)):
                self["_events"][key] = value

        # Call superclass
        dict.__setitem__(self, key, value)


# XXX: There's a problem with attrs accepting EventLists
# so there needs to be a workaround inplace in the future


class EventListMetaclass(type):
    """A Freezeable Metaclass for getting rid of
    the need of freezing different member descriptors"""

    _events: dict[str, event]

    @classmethod
    def __prepare__(
        cls, name: str, bases: tuple[type, ...], /, **kw
    ) -> MutableMapping[str, object]:
        classdict = event_table()
        for b in bases:
            if isinstance(b, EventListMetaclass):
                classdict["_events"].update(b._events)
        return classdict

    def __new__(cls, name: str, bases: tuple[type, ...], classdict: dict, /, **kw):
        return type.__new__(cls, name, bases, classdict, **kw)


class EventList(metaclass=EventListMetaclass):
    """A Subclassable Helper for freezing up multiple callbacks together
    without needing to handle it all yourself"""

    _events: ClassVar[dict[str, event]]

    @cached_property
    def events(self) -> frozenset[str]:
        """An immutable set of event names attached to this class object"""
        return frozenset(self._events.keys())

    def freeze(self):
        """Freezes up all the different callback events
        that were configured"""
        for e in self._events.keys():
            # incase for some reason something is overwritten by the end developer
            object.__getattribute__(self, e).freeze()



import asyncio
import sys
from dataclasses import dataclass
import platform

import pytest
import pytest_asyncio

from aiocallback.events import (EventList, contextevent, event, subclassevent,
                                subcontextevent)

try:
    import attrs
    ATTRS_NOT_FOUND = False
except ModuleNotFoundError:
    ATTRS_NOT_FOUND = True


try:
    # As much as you should try migrating to using msgspec
    # until pydantic decides to migrate it's 
    # metaclass BaseModel to C / Cython / Rust
    # I shall do my fair share of trying to support it.
    import pydantic
    PYDANTIC_NOT_FOUND = False
except ModuleNotFoundError:
    PYDANTIC_NOT_FOUND = True


try:
    # XXX: PyPy does not do well with msgspec apparently...
    if platform.python_implementation() != "PyPy":
        import msgspec
        MSGSPEC_NOT_FOUND = False
    else:
        MSGSPEC_NOT_FOUND = True
except ModuleNotFoundError:
    MSGSPEC_NOT_FOUND = True


attrs_test = pytest.mark.skipif(
    ATTRS_NOT_FOUND, reason="Attrs Not Found!"
)

pydantic_test = pytest.mark.skipif(
    PYDANTIC_NOT_FOUND, reason="Pydantic Not Found!"
)

msgspec_test = pytest.mark.skipif(
    MSGSPEC_NOT_FOUND, reason="Msgspec Not Found!"
)



@pytest.mark.asyncio
async def test_eventlist_freezing():
    
    class MyEvents(EventList):
        @event
        async def my_event_1(self):
            pass
        
        @contextevent
        async def my_event_2(self):
            pass

        @subclassevent
        async def my_event_3(self):
            pass

        @subcontextevent
        async def my_event_4(self):
            pass
    
    my_events = MyEvents()
    my_events.freeze()
    for e in my_events.events:
        assert getattr(my_events, e).frozen, f"{e} did not freeze"


# This is what EventLists were invented for
# - Lazy setting
# - Lazy Initialization
#
# With dataclass support added and now with attrs support, 
# you can get as lazy as you want.

@pytest.mark.asyncio
async def test_dataclass_eventlist_support():
    @dataclass
    class MyEvents(EventList):
        x: int = 0
        y: int = 0
        @event
        async def my_event_1(self):
            pass
        
        @contextevent
        async def my_event_2(self):
            pass

        @subclassevent
        async def my_event_3(self):
            pass

        @subcontextevent
        async def my_event_4(self):
            pass
    
    e = MyEvents(1, 2)
    assert e.x == 1
    assert e.y == 2
    
    for i in range(1, 4):
        assert f"my_event_{i}" in e.events, f"my_event_{i} wasn't registered"
        
    e.freeze()

    # Test Freezing
    for attr in e.events:
        assert e.__getattribute__(attr).frozen, f'{attr} did not freeze'




@attrs_test
@pytest.mark.asyncio
async def test_attrs_eventlist_support():
   
    @attrs.define
    class MyEvents(EventList):
        x: int = 0
        y: int = 0
        @event
        async def my_event_1(self):
            pass
        
        @contextevent
        async def my_event_2(self):
            pass

        @subclassevent
        async def my_event_3(self):
            pass

        @subcontextevent
        async def my_event_4(self):
            pass
    
    e = MyEvents(1, 2)
    assert e.x == 1
    assert e.y == 2
    # Test registration
    for i in range(1, 4):
        assert f"my_event_{i}" in e.events, f"my_event_{i} wasn't registered"
    e.freeze()

    # Test Freezing
    for attr in e.events:
        assert e.__getattribute__(attr).frozen, f'{attr} did not freeze'




@pydantic_test
@pytest.mark.asyncio
async def test_pydantic_eventlist_support():
   
    @pydantic.dataclasses.dataclass
    class MyEvents(EventList):
        x: int = 0
        y: int = 0
        @event
        async def my_event_1(self):
            pass
        
        @contextevent
        async def my_event_2(self):
            pass

        @subclassevent
        async def my_event_3(self):
            pass

        @subcontextevent
        async def my_event_4(self):
            pass
    
    e = MyEvents(1, 2)
    assert e.x == 1
    assert e.y == 2
    # Test registration
    for i in range(1, 4):
        assert f"my_event_{i}" in e.events, f"my_event_{i} wasn't registered"
        
    e.freeze()

    # Test Freezing
    for attr in e.events:
        assert e.__getattribute__(attr).frozen, f'{attr} did not freeze'



@msgspec_test
@pytest.mark.asyncio
async def test_msgspec_struct_support():
    # I don't expect msgspec to support 
    # Eventlists yet since it has strict rules about what goes into 
    # it and out of it.

    # Until Msgspec's StructMeta Class Object is Publically Exposed 
    # I don't Expcet us supporting it anytime soon...

    # However we can test if at least one event will freeze

    class MyEvents(msgspec.Struct, dict=False):
        x:int = 0
        y:int = 0

        @event
        async def my_event_1(self):
            pass

    e = MyEvents(1, 2)
    assert e.x == 1
    assert e.y == 2

    e.my_event_1.freeze()
    assert e.my_event_1.frozen, "msgspec.Struct did not freeze the event"


    




import asyncio
import random
import sys

import pytest
import pytest_asyncio

from aiocallback.events import (EventWrapper, contextevent, event,
                                subclassevent, subcontextevent)




def compute_result():
    """Simple addition function test"""
    a, b = random.choices([1, 2, 3,4, 5, 6, 7, 8, 9, 0], k=2)
    result = a + b
    return result, a, b

def random_value():
    return random.randint(0, 5000)

@pytest.mark.asyncio
async def test_eventwrapper():
    add_event = EventWrapper()
    
    result , a , b = compute_result()

    @add_event
    async def on_add(a:int, b:int):
        nonlocal result
        assert a + b == result, f"on_add did not send {a}+{b} correctly and got {result}"

    add_event.freeze()
    await add_event.send(a, b)


@pytest.mark.asyncio
async def test_eventwrapper_member_descriptor():
    result , a , b = compute_result()

    class TestEvent:
        @event
        async def on_add(self, a:int, b:int) -> None: # type: ignore
            raise RuntimeError("on_add disallows internal calling")
        
    
    test_event = TestEvent()
    
    async def run_test(a:int, b:int):
        nonlocal result
        assert a + b == result, f"on_add event did not send {a}+{b} correctly and got {result}"

    @test_event.on_add
    async def generic_on_add(a:int, b:int):
        await run_test(a, b)

    # Test appending an event
    async def on_add_append(a:int, b:int):
        await run_test(a, b)
    
    test_event.on_add.append(on_add_append)

    assert generic_on_add in test_event.on_add, "generic_on_add was not added"
    assert on_add_append in test_event.on_add, "on_add_append was not added"

    test_event.on_add.freeze()
    assert test_event.on_add.frozen, "on_add is not frozen"
    await test_event.on_add.send(a, b)

@pytest.mark.asyncio
async def test_eventwrapper_subclass_memeber_descriptor():
    
    class TestSubclassEvent:
        def __init__(self):
            self.event_called = False
        
        @subclassevent
        async def subcls_event(self, value:int):
            assert hasattr(self, "event_called"), "TestSubclassEvent Was initalized the wrong way."
            self.event_called = True

    test_sub_event = TestSubclassEvent()

    assert not test_sub_event.event_called, "event_called wasn't called yet but was flagged as true"

    SHOULDVE_CALLED_FOR = random_value()

    @test_sub_event.subcls_event
    async def on_subclass_called(value:int):
        nonlocal SHOULDVE_CALLED_FOR
        assert SHOULDVE_CALLED_FOR == value, "SHOULDVE_CALLED_FOR WAS INCORRECT Value:{!r}".format(value)

    assert on_subclass_called in test_sub_event.subcls_event, "callback wasn't given"

    test_sub_event.subcls_event.freeze()

    await test_sub_event.subcls_event.send(SHOULDVE_CALLED_FOR)

    assert test_sub_event.event_called, "subcls_event Was not Called"

@pytest.mark.asyncio
async def test_eventwrapper_contextevent():
    class TestContextEvent:
        def __init__(self):
            self.event_called = False
        
        @contextevent
        async def subcls_event(self, _:int) -> None: # type: ignore
            raise RuntimeError("class's inner event wasn't ignored")

    
    test_sub_event = TestContextEvent()

    assert not test_sub_event.event_called, "event_called wasn't called yet but was flagged as true"

    SHOULDVE_CALLED_FOR = random_value()

    @test_sub_event.subcls_event
    async def on_subclass_called(event:TestContextEvent, value:int):
        assert hasattr(event, "event_called"), "instanced of the event should've been passed"

        nonlocal SHOULDVE_CALLED_FOR
        assert SHOULDVE_CALLED_FOR == value, "SHOULDVE_CALLED_FOR WAS INCORRECT, \"Value Passed\":{!r}".format(value)

    assert on_subclass_called in test_sub_event.subcls_event, "subclassevent member descriptor had an attribute problem"
    test_sub_event.subcls_event.freeze()
    await test_sub_event.subcls_event.send(SHOULDVE_CALLED_FOR)
    assert not test_sub_event.event_called, "Event Shouldn't have been set to begin with"


@pytest.mark.asyncio
async def test_eventwrapper_subcontextevent():
    class TestSubContextEvent:
        def __init__(self):
            self.event_called = False
        
        @subcontextevent
        async def subcls_event(self, _:int) -> None: # type: ignore
            assert hasattr(self, "event_called"), "TestSubContextEvent Was initalized the wrong way."
            self.event_called = True

    test_sub_event = TestSubContextEvent()

    assert not test_sub_event.event_called, "event_called wasn't called yet but was flagged as true"

    SHOULDVE_CALLED_FOR = random_value()

    @test_sub_event.subcls_event
    async def on_subclass_called(event:TestSubContextEvent, value:int):
        assert hasattr(event, "event_called"), "instance of this event should've been passed"

        nonlocal SHOULDVE_CALLED_FOR
        assert SHOULDVE_CALLED_FOR == value, "SHOULDVE_CALLED_FOR WAS INCORRECT, \"Value Passed\":{!r}".format(value)

    assert on_subclass_called in test_sub_event.subcls_event, "subclassevent member descriptor had an attribute problem"
    test_sub_event.subcls_event.freeze()
    await test_sub_event.subcls_event.send(SHOULDVE_CALLED_FOR)

    assert test_sub_event.event_called, "Inner function was not called"











# ruff: noqa: E402,F401
from .events import (
    EventList,
    EventListMetaclass,
    EventWrapper,
    SelfEventWrapper,
    contextevent,
    event,
    subclasscontextevent,
    subcontextevent,
    subclassevent,
)

__author__ = "Vizonex"

from .__version__ import __version__
